import datetime
from typing import List, cast
import numpy as np
from minder.dataset_manager._utils import Dataset
from minder.dataset_manager import Preprocessor
from minder.research_portal_client.models import ExportJobRequest, ExportJobRequestDataset
from minder.research_portal_client import JobManager


class UnlabelledUtiDataset(Dataset):
    attribute_map = {
        "activity": "activity",
        "patients": "patient_id",
        "dates": "dates",
    }

    def __init__(self, *, job_id: List[str], activity: np.ndarray, patients: np.ndarray, dates: np.ndarray):
        super().__init__(job_id=job_id)
        self.activity = activity
        self.patients = patients
        self.dates = dates

    @classmethod
    def get_latest_data_request(cls, date: datetime.datetime) -> List[ExportJobRequest]:
        today = datetime.datetime.combine(date.date(), datetime.time(0, 0), datetime.timezone.utc)
        last_sunday = today - datetime.timedelta(days=(date.weekday() + 7 + 1) % 7)
        sunday_a_week_ago = last_sunday - datetime.timedelta(days=7)

        return [
            ExportJobRequest(
                since=sunday_a_week_ago,
                until=last_sunday,
                common_observation_columns=["start_date", "patient_id"],
                omit_units=True,
                datasets={
                    "raw_door_sensor": ExportJobRequestDataset(columns=["location_name", "value"]),
                    "raw_appliance_use": ExportJobRequestDataset(columns=["location_name", "value"]),
                    "raw_activity_pir": ExportJobRequestDataset(columns=["location_name"]),
                },
            )
        ]

    @classmethod
    def create(cls, job_id: List[str], files: List[str]):
        preprocessor = Preprocessor(files)

        activity_data_df = preprocessor.load_activity_data()
        activity_data_df = preprocessor.standardise_activity_data(activity_data_df)

        activity_data_nd = preprocessor.reshape_activity_data(activity_data_df)
        _, patients, dates, _ = preprocessor.extract_data_from_index(activity_data_df.index)

        return cls(job_id=job_id, activity=activity_data_nd, patients=patients, dates=dates)

    async def update(self, latest_data: Dataset, *, job_manager: JobManager = None) -> "UnlabelledUtiDataset":
        await super().update(latest_data, job_manager=job_manager)

        latest_data = cast(UnlabelledUtiDataset, latest_data)

        update_start_date: datetime.date = min(latest_data.dates)
        mask = self.dates < update_start_date

        self.filter(mask)

        self.activity = np.concatenate([self.activity, latest_data.activity])
        self.patients = np.concatenate([self.patients, latest_data.patients])
        self.dates = np.concatenate([self.dates, latest_data.dates])

        return self

    def filter(self, mask: np.ndarray) -> "UnlabelledUtiDataset":
        self.activity = self.activity[mask]
        self.patients = self.patients[mask]
        self.dates = self.dates[mask]

        return self



import datetime
from functools import cached_property
from typing import List, cast
from minder.dataset_manager._utils import Dataset, ComposedDataset
from minder.dataset_manager.datasets import UtiLabelsDataset, UnlabelledUtiDataset
from minder.research_portal_client.models import ExportJobRequest
from minder.research_portal_client import JobManager
import pandas as pd


@ComposedDataset.of(activity=UnlabelledUtiDataset, uti_labels=UtiLabelsDataset)
class LabelledUtiDataset(ComposedDataset):
    def __init__(self, *, activity: UnlabelledUtiDataset, uti_labels: UtiLabelsDataset):
        super().__init__(activity=activity, uti_labels=uti_labels)
        mask = ~self.labels.isna()
        activity.filter(mask)
        try:
            delattr(self, "labels")
        except AttributeError:
            pass

    @cached_property
    def labels(self):
        activity: UnlabelledUtiDataset = self["activity"]
        uti_labels: UtiLabelsDataset = self["uti_labels"]
        return pd.MultiIndex.from_arrays([activity.patients, activity.dates]).map(uti_labels.labels)

    @classmethod
    def get_latest_data_request(cls, date: datetime.datetime) -> List[ExportJobRequest]:
        label_request = cls._dataset_cls["label"].get_latest_data_request(date)[0]
        activity_request = cls._dataset_cls["activity"].get_latest_data_request(date)[0]
        activity_request.datasets.update(label_request.datasets)
        activity_request.since = label_request.since
        activity_request.until = label_request.until

        return [activity_request]

    async def update(self, latest_data: Dataset, *, job_manager: JobManager = None) -> "LabelledUtiDataset":
        latest_data = cast(LabelledUtiDataset, latest_data)
        latest_activity = cast(UnlabelledUtiDataset, latest_data["activity"])
        latest_labels = cast(UtiLabelsDataset, latest_data["uti_labels"])

        earliest_label = min(pd.MultiIndex.from_tuples(latest_labels.labels.keys()).get_level_values(1))
        earliest_data = min(latest_activity.dates)

        # Need to download previous job
        if earliest_label < earliest_data:
            if job_manager is None:
                raise ValueError("job_manager not provided")

            previous_job_id = self["activity"]._job_id[-1]
            files = await Dataset.download([previous_job_id], job_manager)
            previous_dataset = UnlabelledUtiDataset.create(previous_job_id, files)

            mask = (previous_dataset.dates >= earliest_label) & (previous_dataset.dates < earliest_data)
            previous_dataset.filter(mask)
            missing_labelled = LabelledUtiDataset(activity=previous_dataset, uti_labels=latest_labels)

            await super().update(missing_labelled, job_manager=job_manager)

        await super().update(latest_data, job_manager=job_manager)

        try:
            delattr(self, "labels")
        except AttributeError:
            pass

        return self



import asyncio
import datetime
from os import PathLike
from typing import Any, Dict, List, Union, final
from abc import ABC, abstractmethod
import numpy as np
from minder.research_portal_client.models import ExportJobRequest
from minder.research_portal_client import JobManager


class Dataset(ABC):
    attribute_map: Dict[str, str] = {}

    def __init__(self, *, job_id: List[str]):
        self.__job_id = job_id

    @final
    @property
    def _job_id(self):
        return self.__job_id

    def to_dict(self):
        result = {}

        for attr, attrKey in self.attribute_map.items():
            if hasattr(self, attr):
                value = getattr(self, attr)
                if value is not None:
                    result[attrKey] = value

        result["__job_id"] = self._job_id

        return result

    @final
    def save(self, output_path: Union[str, PathLike]):
        output = self.to_dict()
        np.savez_compressed(output_path, **output)

    @final
    @classmethod
    def load(cls, path: Union[str, PathLike]):
        with np.load(path, mmap_mode="r", allow_pickle=True) as data:
            kwargs = cls._load(data)

        return cls(**kwargs)

    @classmethod
    def _load(cls, data: Dict[str, Any], prefix: str = "") -> Dict[str, Any]:
        kwargs = {}
        for attr, attrKey in cls.attribute_map.items():
            prefixedKey = f"{prefix}{attrKey}"
            if prefixedKey in data:
                kwargs[attr] = data[prefixedKey]

        kwargs["job_id"] = data[f"{prefix}__job_id"].tolist()

        return kwargs

    @classmethod
    def configure(cls, **kwargs) -> None:
        pass

    @classmethod
    @abstractmethod
    def get_latest_data_request(cls, date: datetime.datetime) -> List[ExportJobRequest]:
        pass

    @final
    @classmethod
    async def download(cls, job_ids: List[str], job_manager: JobManager):
        tasks = []
        for job_id in job_ids:
            tasks.append(job_manager.download(job_id))

        result = await asyncio.gather(*tasks)

        files: List[str] = []
        for f in result:
            files.extend(f)

        return files

    @classmethod
    @abstractmethod
    def create(cls, job_id: List[str], files: List[str]):
        pass

    @abstractmethod
    async def update(self, latest_data: "Dataset", *, job_manager: JobManager = None) -> "Dataset":
        self.__job_id.extend(latest_data.__job_id)

        # Remove duplicates
        self.__job_id = list(dict.fromkeys(self.__job_id))

        return self



import asyncio
import datetime
from typing import Any, Dict, List, Tuple, Type, Union, cast, final
from minder.research_portal_client.models import ExportJobRequest
from minder.research_portal_client import JobManager
from minder.dataset_manager._utils import Dataset


class ComposedDataset(Dataset):
    _dataset_cls: Dict[str, Type[Dataset]] = {}

    def __init__(self, **datasets: Dataset):
        super().__init__(job_id=[])
        self.__datasets = datasets

    def __getitem__(self, key: str) -> Dataset:
        return self.__datasets[key]

    def to_dict(self):
        result = {}

        for name, dataset in self.__datasets.items():
            dataset_dict = dataset.to_dict()
            for k, v in dataset_dict.items():
                result[f"{name}:{k}"] = v

        return result

    @classmethod
    def _load(cls, data: Dict[str, Any], prefix: str = "") -> Dict[str, Any]:
        kwargs = {}
        for name, dataset_cls in cls._dataset_cls.items():
            full_prefix = f"{prefix}{name}:"
            dataset_kwargs = dataset_cls._load(data, prefix=full_prefix)

            kwargs[name] = dataset_cls(**dataset_kwargs)

        return kwargs

    @classmethod
    def configure(cls, **kwargs: Dict[str, Any]) -> None:
        for name, dataset_cls in cls._dataset_cls.items():
            if name in kwargs:
                dataset_cls.configure(**kwargs[name])

    @classmethod
    def get_latest_data_request(cls, date: datetime.datetime) -> List[ExportJobRequest]:
        result: Dict[Tuple[Union[datetime.datetime, None], Union[datetime.datetime, None]], ExportJobRequest] = {}
        for dataset_cls in cls._dataset_cls.values():
            cls_requests = dataset_cls.get_latest_data_request(date)
            for request in cls_requests:
                key = (
                    request.since if hasattr(request, "since") else None,
                    request.until if hasattr(request, "until") else None,
                )
                existing_request = result.get(key)
                if existing_request is not None:
                    existing_request.datasets.update(*request)
                else:
                    result[key] = request

        return list(result.values())

    @classmethod
    def create(cls, job_id: List[str], files: List[str]):
        kwargs: Dict[str, Dataset] = {}
        for name, dataset_cls in cls._dataset_cls.items():
            kwargs[name] = dataset_cls.create(job_id, files)

        return cls(**kwargs)

    async def update(self, latest_data: "Dataset", *, job_manager: JobManager = None) -> "Dataset":
        latest_data = cast(ComposedDataset, latest_data)

        tasks = []
        for name in type(self)._dataset_cls.keys():
            task = self.__datasets[name].update(latest_data.__datasets[name], job_manager=job_manager)
            tasks.append(task)

        await asyncio.gather(*tasks)

        return self

    @final
    @staticmethod
    def of(**kwargs: Type[Dataset]):
        def decorator(cls):
            cls._dataset_cls = kwargs
            return cls

        return decorator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['minder',
 'minder.dataset_manager',
 'minder.dataset_manager._utils',
 'minder.dataset_manager.datasets']

package_data = \
{'': ['*']}

modules = \
['py']
install_requires = \
['minder.research-portal-client>=0.1.3,<0.2.0',
 'numpy>=1.22.4,<2.0.0',
 'pandas>=1.4.2,<2.0.0']

setup_kwargs = {
    'name': 'minder.dataset-manager',
    'version': '0.1.0',
    'description': '',
    'long_description': '# Dataset Manager\n\nLibrary to pre-process CSV files from Research Portal into usable datasets.\n\n## Installation\n\n```bash\npip install minder.dataset-manager\n```\n\n## Example\n```python\nimport logging\nimport asyncio\nimport sys\nfrom typing import Optional\nfrom minder.dataset_manager._utils import Dataset\nfrom minder.dataset_manager.datasets import LabelledUtiDataset\nfrom minder.research_portal_client import Configuration, JobManager\n\n\nlogging.basicConfig(level=logging.INFO)\n\nConfiguration.set_default(\n    Configuration(\n        access_token="---REDACTED---",\n    )\n)\n\n\nasync def example1():\n    job_ids = ["c25249e0-82ff-43d1-9676-f3cead0228b9"]\n    async with JobManager() as job_manager:\n        files = Dataset.download(job_ids, job_manager)\n        dataset = LabelledUtiDataset.create(job_ids, files)\n        dataset.save("./my-dataset.npz")\n\n\nasync def example2():\n    job_ids = ["c25249e0-82ff-43d1-9676-f3cead0228b9"]\n    existing_dataset = "./my-dataset.npz"\n    async with JobManager() as job_manager:\n        download_task = Dataset.download(job_ids, job_manager)\n        try:\n            previous_dataset: Optional[Dataset] = None\n            if existing_dataset.exists():\n                previous_dataset = LabelledUtiDataset.load(existing_dataset)\n        finally:\n            files = await download_task\n\n        new_dataset = LabelledUtiDataset.create(job_ids, files)\n\n        dataset = (\n            await previous_dataset.update(new_dataset, job_manager=job_manager)\n            if previous_dataset is not None\n            else new_dataset\n        )\n        dataset.save("./my-dataset.npz")\n\n\nasync def main():\n    await example1()\n    await example2()\n\n\nif sys.platform == "win32"::\n    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())\n\nasyncio.run(main())\n```\n\n# Development\n\n## Useful commands\n\n### Setup\n\n```bash\npoetry install\n```\n\n### Run tests\n  \n```bash\npoetry run pytest\n```\n\n### Code Coverage\n\nThis command consists of 2 parts:\n- running tests with coverage collection\n- formatting the report: `report` (text to stdout), `xml` (GitLab compatible: cobertura), `html` (visual)\n\n```bash\npoetry run coverage run -m pytest && poetry run coverage report -m\n```\n\n### Linting\n\n```bash\npoetry run flake8\n```\n\n### Formatting\n\n```bash\npoetry run black .\n```\n\n### Type Checking\n\n```bash\npoetry run mypy .\n```\n',
    'author': 'UK DRI Care Research & Technology centre',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import datetime
import logging
from typing import Dict, List, Tuple, Union
import numpy as np
import pandas as pd
import os


logger = logging.getLogger(__name__)


class Preprocessor(object):
    def __init__(self, files: List[str]):
        self.__files = files

    def load_activity_data(self) -> pd.DataFrame:
        dataset_names = ["raw_door_sensor", "raw_appliance_use", "raw_activity_pir"]
        col_filter = ["patient_id", "start_date", "location_name", "value"]

        activity_dataset: pd.DataFrame = None
        for dataset_name in dataset_names:
            dataset = self.__load_dataset(dataset_name)
            dataset = dataset.loc[dataset.start_date != "start_date"]

            if dataset_name == "raw_appliance_use":
                dataset.location_name = dataset.value.apply(lambda x: x.split("-")[0])

            dataset["value"] = 1
            dataset = dataset[col_filter]
            dataset.columns = ["id", "time", "location", "value"]
            dataset.time = pd.to_datetime(dataset.time, utc=True).dt.tz_convert(None)

            activity_dataset = pd.concat([activity_dataset, dataset], ignore_index=True, copy=False)

        return activity_dataset

    def load_uti_labels(self, manual_labels: pd.DataFrame = None) -> Dict[Tuple[str, datetime.date], bool]:
        procedures: pd.DataFrame = self.__load_dataset("procedure")
        if procedures is None:
            raise __PreprocessorError(message="Procedures dataset not present")

        def uti_filter(x):
            texts = [str(x.type), str(x.notes)]
            keywords = ["urine", "urinalysis", "uti"]

            for text in texts:
                for keyword in keywords:
                    if keyword in text:
                        return True

            return False

        procedures.notes = procedures.notes.apply(lambda x: str(x).lower())
        procedures.type = procedures.type.apply(lambda x: str(x).lower())

        mask = procedures.apply(uti_filter, axis=1)
        procedures = procedures.loc[mask, ["patient_id", "start_date", "outcome", "notes"]]
        procedures.columns = ["patient id", "date", "valid", "notes"]

        outcome_mapping: Dict[str, Union[bool, None]] = {
            "http://snomed.info/sct|260385009": False,  # Negative
            "http://snomed.info/sct|10828004": True,  # Positive
            "http://snomed.info/sct|82334004": None,  # Indeterminate
        }

        def map_outcome(x):
            mapped = outcome_mapping.get(x.valid)
            return "negative" not in x.notes if mapped else mapped

        procedures.valid = procedures.apply(map_outcome, axis=1)
        procedures = procedures[["patient id", "date", "valid"]]
        procedures.date = pd.to_datetime(procedures.date).dt.date
        procedures.dropna(inplace=True)

        if manual_labels is not None:
            procedures = pd.concat([procedures, manual_labels], ignore_index=True, copy=False)

        procedures.drop_duplicates(inplace=True)

        procedures.set_index(["patient id", "date"], inplace=True)
        return procedures["valid"].to_dict()

    def expand_uti_labels(
        self,
        labels: Dict[Tuple[str, datetime.date], bool],
        days_either_side: int = 0,
    ) -> Dict[Tuple[str, datetime.date], bool]:
        if days_either_side == 0:
            return labels

        label_df = pd.DataFrame(
            labels.values(),
            index=pd.MultiIndex.from_tuples(labels.keys(), names=["patient_id", "date"]),
            columns=["valid"],
        )
        label_df.reset_index(inplace=True)

        def dates_either_side_group_by(x):
            date = pd.to_datetime(x["date"].values[0])
            x = [x] * (2 * days_either_side + 1)
            new_date_values = np.arange(-days_either_side, days_either_side + 1)
            new_dates = [(date + datetime.timedelta(int(value))).date() for value in new_date_values]
            x = pd.concat(x)
            x["date"] = new_dates
            return x

        return (
            label_df.groupby(["patient_id", "date", "valid"])
            .apply(dates_either_side_group_by)
            .reset_index(drop=True)
            .set_index(["patient_id", "date"])["valid"]
            .to_dict()
        )

    def apply_uti_labels(
        self,
        unlabelled_df: pd.DataFrame,
        uti_labels: Dict[Tuple[str, datetime.date], bool],
    ) -> pd.DataFrame:
        dropIndex = unlabelled_df.index.names != ["id", "date"]
        if dropIndex:
            dropDate = "date" not in unlabelled_df.columns
            if dropDate:
                unlabelled_df["date"] = unlabelled_df.time.dt.date

            unlabelled_df.set_index(["id", "date"], drop=False, inplace=True)

        unlabelled_df["valid"] = unlabelled_df.index.map(uti_labels)

        if dropIndex:
            unlabelled_df.reset_index(drop=True, inplace=True)

            if dropDate:
                unlabelled_df.drop(columns=["date"], inplace=True)

        return unlabelled_df

    def standardise_activity_data(
        self,
        df: pd.DataFrame,
        *,
        uti_labels: Dict[Tuple[str, datetime.date], bool] = None,
        drop_time: bool = True,
    ) -> pd.DataFrame:
        df.drop_duplicates(inplace=True, ignore_index=True)
        df.loc[:, "date"] = df.time.dt.date

        df_start = df[["id", "date", "location"]].drop_duplicates()
        df_start["time"] = pd.to_datetime(df_start.date)
        df_end = df_start.copy()
        df_end.time += datetime.timedelta(hours=23)

        df = pd.concat([df, df_start, df_end], sort=False, ignore_index=True, copy=False)
        df.drop_duplicates(subset=["id", "time", "location"], inplace=True)
        df.fillna(0.0, inplace=True)

        df = df.groupby(["id", "location"]).apply(lambda x: x.set_index("time").resample("H").sum()).reset_index()
        table_df = df.pivot_table(index=["id", "time"], columns="location", values="value").reset_index()
        table_df.fillna(0.0, inplace=True)

        sensors = [
            "back door",
            "bathroom1",
            "bedroom1",
            "dining room",
            "fridge door",
            "front door",
            "hallway",
            "kettle",
            "kitchen",
            "living room",
            "lounge",
            "microwave",
            "study",
            "toaster",
        ]
        for sensor in sensors:
            if sensor not in table_df.columns:
                table_df[sensor] = 0.0

        table_df.dropna(inplace=True)
        table_df["date"] = table_df.time.dt.date
        table_df = table_df[["id", "date"] + (["time"] if not drop_time else []) + sensors]
        table_df.set_index(["id", "date"], inplace=True)

        if uti_labels is not None:
            table_df = self.apply_uti_labels(table_df, uti_labels)
            table_df.set_index(["valid"], append=True, inplace=True)

        return table_df

    def reshape_activity_data(self, df: pd.DataFrame) -> np.ndarray:
        if "time" in df.columns:
            df.drop(columns="time", inplace=True)

        return df.to_numpy().reshape(-1, 24, len(df.columns))

    def extract_data_from_index(
        self, index: pd.MultiIndex
    ) -> Tuple[pd.MultiIndex, np.ndarray, np.ndarray, Union[np.ndarray, None]]:
        """
        Extract values for each index level.

        Parameters
        ----------
        index : MultiIndex from a standartised datasest. Optianally with labels applied to it.

        Returns
        -------
        Tuple(indices, patient_ids, dates, labels)

        Labels will be None if index did not contain labels.
        """
        indices: pd.MultiIndex = index.drop_duplicates()
        patient_ids = indices.get_level_values(0).astype(str).to_numpy()
        dates = indices.get_level_values(1).to_numpy()
        labels = indices.get_level_values(2).astype(int).to_numpy() if indices.nlevels == 3 else None
        return (indices, patient_ids, dates, labels)

    def split_labeled(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Split the DataFrame into labelled and unlabelled data.

        Split the DataFrame based on 3rd level of its index:
            - `isna() == True` will be in unlabelled data frame
            - the rest will be in the labelled data frame

        Parameters
        ----------
        df : DataFrame with a standartised datasest. Optianally with labels applied to it.
            If no labels are applied, entire DataFrame is treated an unlabelled.

        Returns
        -------
        Tuple of DataFrame or None with 2 elements: labelled and unlabelled.
        """
        if df.index.nlevels != 3:
            return (None, df)

        groups = df.groupby(df.index.get_level_values(2).isna())
        split = [x for _, x in groups]
        if len(split) == 2:
            return (split[0], split[1])

        return (None, split[0]) if pd.isna(df.index[0][2]) else (split[0], None)

    def __load_dataset(self, dataset: str) -> pd.DataFrame:
        concat_csv: pd.DataFrame = None
        for file_path in self.__files:
            filename = os.path.basename(file_path)
            file_dataset = filename.split("-")[0]
            if file_dataset != dataset:
                continue

            csv_content = pd.read_csv(file_path, index_col=False, memory_map=True)
            concat_csv = pd.concat([concat_csv, csv_content], ignore_index=True, copy=False)

        if concat_csv is not None:
            concat_csv.drop_duplicates(inplace=True, ignore_index=True)

        return concat_csv


class __PreprocessorError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        """Custom error messages for exception"""
        return self.message



import datetime
from typing import Dict, List, Tuple, cast
from minder.dataset_manager._utils import Dataset
from minder.dataset_manager import Preprocessor
from minder.research_portal_client.models import ExportJobRequest, ExportJobRequestDataset
from minder.research_portal_client import JobManager
import numpy as np


class UtiLabelsDataset(Dataset):
    attribute_map = {
        "labels": "label",
    }

    __expand_labels: int = 1

    def __init__(self, *, job_id: List[str], labels: np.ndarray):
        super().__init__(job_id=job_id)
        self.labels = cast(Dict[Tuple[str, datetime.date], bool], labels.flat[0])

    @classmethod
    def configure(cls, **kwargs) -> None:
        cls.__configure(**kwargs)

    @classmethod
    def __configure(cls, *, expand_labels: int):
        cls.__expand_labels = expand_labels

    @classmethod
    def get_latest_data_request(cls, date: datetime.datetime) -> List[ExportJobRequest]:
        today = datetime.datetime.combine(date.date(), datetime.time(0, 0), datetime.timezone.utc)
        last_sunday = today - datetime.timedelta(days=(date.weekday() + 7 + 1) % 7)
        sunday_a_week_ago = last_sunday - datetime.timedelta(days=7)
        sunday_2_weeks_ago = sunday_a_week_ago - datetime.timedelta(days=7)

        return [
            ExportJobRequest(
                since=sunday_2_weeks_ago,
                until=sunday_a_week_ago,
                datasets={
                    "procedure": ExportJobRequestDataset(
                        columns=["start_date", "type", "patient_id", "outcome", "notes"]
                    )
                },
            )
        ]

    @classmethod
    def create(cls, job_id: List[str], files: List[str]):
        preprocessor = Preprocessor(files)
        labels = preprocessor.load_uti_labels()
        labels = preprocessor.expand_uti_labels(labels, cls.__expand_labels)

        return cls(job_id=job_id, labels=np.asarray(labels))

    async def update(self, latest_data: "Dataset", *, job_manager: JobManager = None) -> "UtiLabelsDataset":
        await super().update(latest_data, job_manager=job_manager)

        latest_data = cast(UtiLabelsDataset, latest_data)

        self.labels.update(latest_data.labels)

        return self
